
local notif = CreateFrame("Text")
notif:SetSize(40)
notif:SetColor(1, 0, 0)
notif:SetLayer(100)
notif:SetPoint("CENTER", UIParent, "CENTER")

local notify_warn = ""
local notifications = {}
local function err(text)
  notifications = {}
  table.insert(notifications, {text})
end

notif.Tick = coroutine.wrap(function ()
  while true do
    while #notifications == 0 do
      notif:SetText(notify_warn)
      notif:SetColor(1, 0.5, 0.5)
      coroutine.yield()
    end
    
    local nt = table.remove(notifications, 1)
    
    notif:SetText(nt[1])
    
    notif:SetColor(1, 0.5, 0.5)
    notif:SetBorder(true)
    
    coroutine.pause(60)
    
    for i = 1, 60 do
      notif:SetColor(1, 0.5, 0.5, (60 - i) / 60)
      coroutine.yield()
    end
  end
end)



local bloop = CreateFrame("Frame")
bloop:SetBackgroundColor(16 / 255, 89 / 255, 156 / 255)
bloop:SetHeight(768)
bloop:SetWidth(768)
bloop:SetPoint("TOPRIGHT", UIParent, "TOPRIGHT")
bloop:SetLayer(-1)

local width = 20
local height = 20

bloop:SetCoordinateScale(width / 2, height / 2, height * 1.05)

local outer = CreateFrame("Frame")
outer:SetBackgroundColor(16 / 255 * 0.8, 89 / 255 * 0.8, 156 / 255 * 0.8)
outer:SetPoint("TOPLEFT", UIParent, "TOPLEFT")
outer:SetPoint("BOTTOMRIGHT", bloop, "BOTTOMLEFT")
outer:SetLayer(-1)

local zone = {}
local doors = {}
local racks = {}
local roomcounts = {}

local function gitem(x, y)
  if not zone[x] then
    return nil
  end
  
  return zone[x][y]
end

local function sitem(x, y, z)
  if not zone[x] then
    zone[x] = {}
  end
  
  if gitem(x, y) then
    roomcounts[gitem(x, y)] = roomcounts[gitem(x, y)] - 1
  end
  
  if z then
    roomcounts[z] = (roomcounts[z] or 0) + 1
  end
  
  zone[x][y] = z
end

local function toggle_door(x, y)
  if not doors[x] then doors[x] = {} end
  doors[x][y] = not doors[x][y]
end
local function get_door(x, y)
  if not doors[x] then return false end
  return doors[x][y]
end

local function toggle_rack(x, y)
  if not racks[x] then racks[x] = {} end
  racks[x][y] = not racks[x][y]
end
local function get_rack(x, y)
  if not racks[x] then return false end
  return racks[x][y]
end


local function desired_wall(ax, ay, bx, by)
  local ai, bi = gitem(ax, ay), gitem(bx, by)
  local ad, bd = get_door(ax, ay), get_door(bx, by)
  
  if not ai and not bi then return "open" end
  
  if ai and bi and ai ~= bi and not ad ~= not bd then return "door" end
  
  if (ai or bi) and ai ~= bi then return "solid" end
  
  return "empty"
end


local current_room = 1

for i = 17, 20 do
  for j = 1, width do
    sitem(j, i, -1) -- the road
  end
end


local flood = {}
local function flood_get(i, j)
  if not flood[i] then return false end
  return flood[i][j]
end
local function flood_set(i, j, v)
  if not flood[i] then flood[i] = {} end
  flood[i][j] = v
end
local function flood_reset()
  flood = {}
end
local function flood_test(i, j)
  return flood_get(i, j)
end
local function flood_fill(i, j, roomid)
  if flood_get(i, j) or gitem(i, j) ~= roomid then return end
  flood_set(i, j, true)
  flood_fill(i + 1, j, roomid)
  flood_fill(i - 1, j, roomid)
  flood_fill(i, j + 1, roomid)
  flood_fill(i, j - 1, roomid)
end
local dx = {1, -1, 0, 0}
local dy = {0, 0, 1, -1}
local function flood_fill_doors(i, j)
  if flood_get(i, j) then return end
  flood_set(i, j, true)
  
  local titem = gitem(i, j)
  assert(titem)
  for e = 1, 4 do
    local aitem = gitem(i + dx[e], j + dy[e])
    if aitem == titem or (aitem and not get_door(i, j) ~= not get_door(i + dx[e], j + dy[e])) then
      flood_fill_doors(i + dx[e], j + dy[e])
    end
  end
end


local function contiguous(roomid)
  flood_reset()
  local found = false
  for i = 1, 20 do
    for j = 1, 20 do
      if gitem(i, j) == roomid then
        if not found then
          flood_fill(i, j, roomid)
          found = true
        else
          if not flood_test(i, j) then
            return false
          end
        end
      end
    end
  end
  
  return true
end


local function kliked_set(x, y)
  local old = gitem(x, y)
  
  local did = false
  
  if gitem(x, y) == current_room then
    did = true
    sitem(x, y, nil)
  elseif gitem(x, y) == nil then
    did = true
    sitem(x, y, current_room)
  end
  
  if not contiguous(current_room) then
    sitem(x, y, old)
    err("Rooms must remain connected to themselves!")
  else
    if did then
      PlaySound("etch")
    end
  end
end
local kliked = kliked_set


for i = 1, width + 1 do
  for j = 1, height + 1 do
    local fram = CreateFrame("Button", bloop)
    local zon = CreateFrame("Frame", fram)
    zon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", i - 1, j - 1)
    zon:SetWidth(1)
    zon:SetHeight(1)
    --zon:SetBackgroundColor(math.random(), math.random(), math.random(), math.random())
    
    
    if i <= width and j <= height then
      function fram:Click()
        kliked(i, j)
      end
    end
  end
end

local function makebutton(title, ypos)
  local button = CreateFrame("Button", outer)
  button:SetPoint("CENTER", outer, 0.5, 0, 0, ypos)
  button:SetBackgroundColor(0, 0, 0, 0.2)
  
  local tex = CreateFrame("Text", button)
  tex:SetText(title)
  tex:SetSize(30)
  tex:SetColor(1, 1, 1)
  tex:SetPoint("CENTER", button, "CENTER")
  button:SetPoint("TOPLEFT", tex, "TOPLEFT", -12, -6)
  
  return button
end

local croom = CreateFrame("Text", outer)
croom:SetColor(1, 1, 1)
croom:SetSize(30)
croom:SetPoint("CENTER", outer, 0.5, 0, 0, 300)
local function update_editing()
  if current_room == "doors" then
    croom:SetText("Modifying doors")
  elseif current_room == "racks" then
    croom:SetText("Modifying racks")
  elseif current_room then
    croom:SetText("Editing room #" .. current_room)
  else
    croom:SetText("Choosing room to edit")
  end
end
update_editing()

local finishroom = makebutton("New room", 30)
function finishroom:Click()
  current_room = 1
  while roomcounts[current_room] and roomcounts[current_room] > 0 do
    print(current_room, roomcounts[current_room])
    current_room = current_room + 1
  end
  
  kliked = kliked_set
  update_editing()
end

local finishitall = makebutton("Finish mall", 730)

local total_avail_cash = 500000

function finishitall:Click()
  local has_room = false
  
  for i = 1, #roomcounts do
    if roomcounts[i] > 0 then
      has_room = true
    end
    if not contiguous(current_room) then
      -- this should never happen, but
      err("Rooms must be connected to themselves!")
      return
    end
  end
  
  flood_reset()
  flood_fill_doors(20, 20)
  
  for x = 1, 20 do
    for y = 1, 20 do
      if gitem(x, y) and not flood_get(x, y) then
        err("People can't get to parts of your mall! Add doors! Connect to the street!")
        return
      end
    end
  end
  
  if total_avail_cash < get_total_cashola() then
    err("You can't spend that much money!")
    return
  end
  
  if not has_room then
    err("You need to build a mall!")
    return
  end
  
  local built_racks, built_rooms, built_doors = 0, 0, 0
  
  local ites = {}
  for x = 1, width + 1 do
    for y = 1, 17 do
      if gitem(x, y) and get_rack(x, y) then
        built_racks = built_racks + 1
      end
      if gitem(x, y) then
        ites[gitem(x, y)] = (ites[gitem(x, y)] or 0) + 1
      end
      
      if x <= width then
        -- do what's above us
        if desired_wall(x, y, x, y - 1) == "door" then
          built_doors = built_doors + 1
        end
      end
      
      if y <= 16 then
        -- do what's to the left us
        if desired_wall(x, y, x - 1, y) == "door" then
          built_doors = built_doors + 1
        end
      end
    end
  end

  for k, v in pairs(ites) do
    built_rooms = built_rooms + 1
  end
  built_rooms = built_rooms - 1 -- compensate for -1
  
  results = {zone = zone, doors = doors, racks = racks, built_racks = built_racks, built_rooms = built_rooms, built_doors = built_doors}
end


local editroom = makebutton("Edit room", 80)

function editroom:Click()
  print("eroom")
  current_room = nil
  kliked = function (x, y)
    local ite = gitem(x, y)
    if ite then
      current_room = ite
      kliked = kliked_set
      update_editing()
    end
  end
  
  update_editing()
end


local doors = makebutton("Modify doors", 130)

function doors:Click()
  current_room = "doors"
  
  kliked = function (x, y)
    toggle_door(x, y)
    PlaySound("etch")
  end
  
  update_editing()
end


local racks = makebutton("Modify racks", 180)

function racks:Click()
  current_room = "racks"
  
  kliked = function (x, y)
    toggle_rack(x, y)
    PlaySound("etch")
  end
  
  update_editing()
end



local bloopwhite = CreateFrame("Frame", bloop)

local thinnity = 0.5



function bloopwhite:Render()
  gl.Color(1, 1, 1)
  
  for typ = 1, 4 do
    local tid
    if typ == 1 then
      tid = "solid"
    elseif typ == 2 then
      tid = "open"
    elseif typ == 3 then
      tid = "door"
    elseif typ == 4 then
      tid = "rack"
    end
    local tex = Texture(tid)
    tex:SetTexture()
    
    local tx, ty = tex:GetWidth() / tex:GetInternalWidth(), tex:GetHeight() / tex:GetInternalHeight()
    
    local coord = {}
    
    for x = 1, width + 1 do
      for y = 1, 17 do
        if tid == "rack" then
          if gitem(x, y) and get_rack(x, y) then
            table.insert(coord, x - 1)
            table.insert(coord, y - 1)
            table.insert(coord, x)
            table.insert(coord, y - 1)
            table.insert(coord, x)
            table.insert(coord, y)
            table.insert(coord, x - 1)
            table.insert(coord, y)
          end
        end
        
        if x <= width then
          -- do what's above us
          if desired_wall(x, y, x, y - 1) == tid then
            table.insert(coord, x - 2)
            table.insert(coord, y - 1 - thinnity)
            table.insert(coord, x - 2)
            table.insert(coord, y- 1 + thinnity)
            table.insert(coord, x + 1)
            table.insert(coord, y - 1 + thinnity)
            table.insert(coord, x + 1)
            table.insert(coord, y - 1 - thinnity)
          end
        end
        
        if y <= 16 then
          -- do what's to the left us
          if desired_wall(x, y, x - 1, y) == tid then
            table.insert(coord, x - 1 - thinnity)
            table.insert(coord, y - 2)
            table.insert(coord, x- 1 + thinnity)
            table.insert(coord, y - 2)
            table.insert(coord, x- 1 + thinnity)
            table.insert(coord, y + 1)
            table.insert(coord, x - 1 - thinnity)
            table.insert(coord, y + 1)
          end
        end
      end
    end
    
    local tex = {}
    while #tex < #coord do
      table.insert(tex, 0)
      table.insert(tex, 0)
      table.insert(tex, tx)
      table.insert(tex, 0)
      table.insert(tex, tx)
      table.insert(tex, ty)
      table.insert(tex, 0)
      table.insert(tex, ty)
    end
    
    glutil.RenderArray("QUADS", 2, coord, nil, nil, 2, tex)
  end
  

  local tex = Texture("road")
  tex:SetTexture()
  
  local tx, ty = tex:GetWidth() / tex:GetInternalWidth(), tex:GetHeight() / tex:GetInternalHeight()
  
  local vert = {}
  local tex = {}
  
  for i = -1, width - 2, 4 do
    table.append(vert, {i, 15, i + 6, 15, i + 6, 21, i, 21})
    table.append(tex, {0, 0, tx, 0, tx, ty, 0, ty})
  end
  
  glutil.RenderArray("QUADS", 2, vert, nil, nil, 2, tex)
  
  SetNoTexture()
end

local normal_wall = 500
local door = 5000
local flooring = 1000
local rack = 10000

local moneybox = CreateFrame("Frame", outer)
moneybox:SetPoint("TOPLEFT", outer, "TOPLEFT", 30, 400)
moneybox:SetPoint("RIGHT", outer, "RIGHT", -30)
moneybox:SetBackgroundColor(0, 0, 0, 0.05)

local hand = CreateFrame("Frame", moneybox)
hand:SetPoint("BOTTOMLEFT", moneybox, "TOPLEFT", 20, 10)

local function moneytag(name, amount, ypos)
  local tag = CreateFrame("Text", moneybox)
  tag:SetText(name .. ": $" .. amount)
  tag:SetColor(1, 1, 1)
  tag:SetSize(30)
  tag:SetPoint("TOPLEFT", hand, "BOTTOMLEFT", 0, 10)
  hand = tag
end
moneytag("Floor", flooring)
moneytag("Wall", normal_wall)
moneytag("Door", door)
moneytag("Rack", rack)

moneybox:SetPoint("BOTTOM", hand, "BOTTOM", 0, 20)

function get_total_cashola()
  local cost = -(20 * normal_wall + 20 * flooring)
  
  local function add_wall(wall)
    if wall == "door" then
      cost = cost + door
    end
    if wall == "solid" then
      cost = cost + normal_wall
    end
  end
  
  for x = 1, width + 1 do
    for y = 1, 17 do
      if gitem(x, y) then
        cost = cost + flooring
        
        if get_rack(x, y) then
          cost = cost + rack
        end
      end
      
      if x <= width then
        add_wall(desired_wall(x, y, x, y - 1))
      end
      
      if y <= 16 then
        add_wall(desired_wall(x, y, x - 1, y))
      end
    end
  end
  
  return cost
end

local tag = CreateFrame("Text")
tag:SetText("Money left:")
tag:SetColor(1, 1, 1)
tag:SetSize(30)
tag:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", 20, -100)

local dun = CreateFrame("Text")
dun:SetColor(1, 1, 1)
dun:SetSize(30)
dun:SetPoint("TOP", tag, "TOP")
dun:SetPoint("RIGHT", outer, "RIGHT", -20)
function dun:Tick()
  self:SetText("$" .. tostring(total_avail_cash - get_total_cashola()))
  if total_avail_cash - get_total_cashola() < 0 then
    self:SetColor(1, 0.5, 0.5)
  else
    self:SetColor(1, 1, 1)
  end
end

dun:Tick()
